import { isPlainObject } from '@vue-formily/util';
import { cascadeRules, normalizeRules, getSchemaAcceptance, invalidateSchemaValidation, acceptSchema, genFields } from '../../helpers';
import Element from './Element';
import { logMessage, readonlyDumpProp, readonlyDef } from '../../utils';
import Validation from '../validations/Validation';
const FORM_TYPE = 'group';

async function onFieldChanged(...args) {
  const [field] = args;
  const {
    valid,
    model
  } = field;
  this._d.pending = true;

  if (valid) {
    let value = this._d.value;

    if (!value) {
      value = this._d.value = {};
    }

    readonlyDef(value, model, () => this[model].value);
  }

  if (this.options.silent) {
    await this.validate({
      cascade: false
    });
  }

  this.emit('changed', this, ...args);
}

export default class Group extends Element {
  constructor(schema, parent) {
    super(schema, parent);
    const accepted = Group.accept(schema);

    if (!accepted.valid) {
      throw new Error(logMessage(`invalid schema, ${accepted.reason}`, accepted.infos));
    }

    readonlyDumpProp(this, 'formType', FORM_TYPE);
    readonlyDumpProp(this, 'type', 'enum');

    if (schema.rules) {
      schema.fields = cascadeRules(schema.rules, schema.fields);
    }

    this.fields = genFields(schema.fields, this);
    this.fields.forEach(field => {
      this[field.model] = field;
      field.on('changed:formy', (...args) => onFieldChanged.apply(this, args), {
        noOff: true
      });
    });
    this._d.validation = new Validation(normalizeRules(schema.rules, this.type));
    this._d.value = null;
  }

  static accept(schema) {
    const {
      accepted,
      sv
    } = getSchemaAcceptance(schema, FORM_TYPE);

    if (!accepted) {
      const {
        fields,
        formId
      } = schema;

      if (!Array.isArray(fields)) {
        invalidateSchemaValidation(sv, '`fields` must be an array.', {
          formId
        });
      }

      if (sv.valid) {
        acceptSchema(schema, FORM_TYPE);
      }
    }

    return sv;
  }

  static create(schema, parent) {
    return new Group(schema, parent);
  }

  get pending() {
    return this._d.pending;
  }

  get value() {
    return this._d.value;
  }

  async setValue(obj) {
    if (!isPlainObject(obj)) {
      throw new Error(logMessage('Invalid value, Group value must be an object'));
    }

    await Promise.all(Object.keys(obj).map(async model => {
      const field = this[model];

      if (field) {
        await field.setValue(obj[model]);
      }
    }));
    return this.value;
  }

  shake({
    cascade = true
  } = {}) {
    super.shake();

    if (cascade) {
      this.fields.forEach(field => field.shake());
    }
  }

  isValid() {
    return this.validation.valid && !this.fields.some(field => !field.valid);
  }

  reset() {
    this.cleanUp();
    this.fields.forEach(field => field.reset());
    this.validation.reset();
  }

  async clear() {
    this.cleanUp();
    await Promise.all(this.fields.map(async field => await field.clear()));
  }

  async validate({
    cascade = true
  } = {}) {
    this.emit('validate', this);

    if (cascade) {
      await Promise.all(this.fields.filter(field => 'validate' in field).map(async field => await field.validate()));
    }

    await this.validation.validate(this.value, {}, this.props, this);

    if (!this.valid) {
      this._d.value = null;
    }

    this._d.pending = false;
    this.emit('validated', this);
  }

}
Group.FORM_TYPE = FORM_TYPE;